use anyhow::anyhow;
use rppal::{
    gpio::{Gpio, OutputPin},
    spi::{Bus, Mode, SlaveSelect, Spi},
};
use std::{thread::sleep, time::Duration};

use crate::packet::FecPacket;

const SEND_PACKET_CMD: u8 = 0x01;
const NOP_CMD: u8 = 0x55;

const PREAMBLE: u8 = 0xE5;
const NO_SPACE: u8 = 0x55;
const FREE_SPACE: u8 = 0xAA;

// Used for talking to an RF4463 via a microcontroller
// The microcontroller takes in packet data over SPI and sends it to the RF4463
pub struct McuRadio {
    spi: Spi,
    reset_pin: OutputPin,
}

impl McuRadio {
    pub fn new() -> anyhow::Result<Self> {
        let spi = Spi::new(Bus::Spi0, SlaveSelect::Ss1, 1_000_000, Mode::Mode0)?;
        let reset_pin = Gpio::new()?.get(8)?.into_output_high();

        let mut s = Self { reset_pin, spi };

        s.reset();

        Ok(s)
    }

    pub fn reset(&mut self) {
        self.reset_pin.set_low();
        sleep(Duration::from_millis(10));
        self.reset_pin.set_high();
        sleep(Duration::from_millis(1000));
    }

    pub fn send_packet(&mut self, packet: &FecPacket) -> anyhow::Result<()> {
        // wait until we have packet space

        while !self.get_info()?.0 {
            sleep(Duration::from_millis(10));
        }

        self.spi.write(&[SEND_PACKET_CMD])?;
        self.spi.write(&packet.0)?;

        Ok(())
    }

    pub fn get_temp(&mut self) -> anyhow::Result<f32> {
        Ok(self.get_info()?.1)
    }

    // try 10 times and then give up
    fn get_info(&mut self) -> anyhow::Result<(bool, f32)> {
        for _ in 0..9 {
            if let Ok(x) = self.get_info_once() {
                return Ok(x);
            }
        }

        self.get_info_once()
    }

    fn get_info_once(&mut self) -> anyhow::Result<(bool, f32)> {
        // get 256 bytes from the mcu
        let mut buf = [0; 256];
        self.spi.transfer(&mut buf, &[NOP_CMD; 256])?;

        // look for our data
        // it goes [0xE5, 0xE5, 0xAA || 0x55, ?, ?, ?, ?, 0xE5, 0xE5]
        for i in 0..(buf.len() - 9) {
            let chunk = &buf[i..(i + 9)];

            if chunk[0..2] == [PREAMBLE, PREAMBLE]
                && (chunk[2] == FREE_SPACE || chunk[2] == NO_SPACE)
                && chunk[7..9] == [PREAMBLE, PREAMBLE]
            {
                let free = chunk[2] == FREE_SPACE;
                let temp = f32::from_le_bytes(chunk[3..7].try_into().unwrap());

                return Ok((free, temp));
            }
        }

        Err(anyhow!("could not find valid data"))
    }
}
